#include "include.h"

void Engine::init(void *param1, void *param2)
{
#ifdef WIN32
	hwnd = *((HWND *)param1);
	hdc = *((HDC *)param2);
	gfx.init(param1, param2);

#endif

#ifdef __linux__
	display = (Display *)param1;
	window = *((Window *)param2);
#endif

#ifdef __MACH__
	window = *((WindowPtr *)param1);
	agl_context = *((AGLContext *)param2);
#endif


	zNear = 1;
	zFar  = 1001;
	glClearColor(0, 0, 0.9f, 1.0f);
	glClearStencil(0);

	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);
	glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_STENCIL_TEST);
//	glEnable(GL_STENCIL_TEST_TWO_SIDE_EXT);


	light_frame.reset();
	light_frame.pos = vec3(-32.0f, 12.0f, 0.0f);
	light_frame.forward = vec3(-0.8454f, 0.3135f, 0.4286f);
	light_frame.forward.normalize();
	light_frame.up = vec3(0.2796f, 0.9478f, -0.1418f);
	light_frame.up.normalize();

	camera_frame.reset();
	camera_frame.pos = vec3(-3.0f, -1.0f, -7.0f);

//	model.load(gfx, "bunny3");
	projection.perspective(45.0, (float)width / height, zNear, zFar, false);
	shader.init(&gfx);

}

void Engine::draw_plane(plane_t &p, bool lights)
{
	vec3 color(0.0f, 0.3f, 0.0f);
	matrix4	mat;
	vec3	right(1.0f, 0.0f, 0.0f);
	vec3	forward(0.0f, 0.0f, 1.0f);
	float size = 500.0f;

	if (lights == false)
	{
		color = vec3(0.0f, 0.1f, 0.0f);
	}

	mat.m[0] = right.x;
	mat.m[1] = right.y;
	mat.m[2] = right.z;
	mat.m[3] = 0.0f;

	mat.m[4] = p.normal.x;
	mat.m[5] = p.normal.y;
	mat.m[6] = p.normal.z;
	mat.m[7] = 0.0f;

	mat.m[8] = forward.x;
	mat.m[9] = forward.y;
	mat.m[10] = forward.z;
	mat.m[11] = 0.0f;

	mat.m[12] = -(right * vec3(0.0f, 0, 0.0f));
	mat.m[13] = -(p.normal * vec3(0.0f, p.d, 0.0f));
	mat.m[14] = -(forward * vec3(0.0f, 0, 0.0f));
	mat.m[15] = 1.0f;


	matrix4 mvp = transformation.premultiply(mat.m) * projection;
	shader.Params(mvp, color);
	glBegin(GL_QUADS);

	//frontface
	glColor3f(color.x, color.y, color.z);
	glTexCoord2f(1.0f, 1.0f);
	glVertex3f(size, 0.0f, size);
	glTexCoord2f(1.0f, 0.0f);
	glVertex3f(size, 0.0f, -size);
	glTexCoord2f(0.0f, 0.0f);
	glVertex3f(-size, 0.0f, -size);
	glTexCoord2f(0.0f, 1.0f);
	glVertex3f(-size, 0.0f, size);

	//backface
	glTexCoord2f(1.0f, 1.0f);
	glVertex3f(size, 0.0f, size);
	glTexCoord2f(0.0f, 1.0f);
	glVertex3f(-size, 0.0f, size);
	glTexCoord2f(0.0f, 0.0f);
	glVertex3f(-size, 0.0f, -size);
	glTexCoord2f(1.0f, 0.0f);
	glVertex3f(size, 0.0f, -size);

	glEnd();

}

void Engine::render()
{
	const float identity[16] = {
		1.0f, 0.0f, 0.0f, 0.0f,
		0.0f, 1.0f, 0.0f, 0.0f,
		0.0f, 0.0f, 1.0f, 0.0f,
		0.0f, 0.0f, 0.0f, 1.0f };


	if (keyboard.control)
		light_frame.update(keyboard);
	else
		camera_frame.update(keyboard);


	shader.Select();
	gfx.clear();
	transformation = identity;

	if (keyboard.control)
		light_frame.set(transformation);
	else
		camera_frame.set(transformation);

	render_scene(true);
	gfx.swap();
	
}

void Engine::render_scene(bool lights)
{
	vec3 color = vec3(0.0, 1.0, 0.0);
	plane_t floor = { {0.0f, 1.0f, 0.0f}, 3.0f };

	draw_plane(floor, lights);

	for(unsigned int i = 0; i < entity_list.size(); i++)
	{
		if (lights)
		{
			color = vec3(1.0f, 1.0f, 1.0f);
		}
		else
		{
			color = vec3(0.5f, 0.5f, 0.5f);
		}


		matrix4 matrix;

		entity_list[i]->get_matrix(matrix.m);

		matrix4 mvp = transformation.premultiply(matrix.m) * projection;
		color.x = 0;
		color.y = 0;
		color.z = 0;
		shader.Params(mvp, color);
		model.render(gfx);

	}

}

void Engine::dynamics()
{
	cfg_t	config;
	for(unsigned int i = 0; i < entity_list.size(); i++)
	{
		if (entity_list[i]->sleep == true )
			continue;

		RigidBody *body = entity_list[i];
		float delta_time = 0.016f;
		float target_time = delta_time;
		float current_time = 0.0f;
		int divisions = 0;

		if (body->gravity)
			body->net_force = vec3(0.0f, -9.8f * body->mass, 0.0f);

		while (current_time < delta_time)
		{
			body->save_config(config);
			body->integrate(target_time - current_time);
			if ( collision_detect(*body) )
			{
				body->load_config(config);
				target_time = (current_time + target_time) / 2.0f;
				divisions++;

				if (divisions > 10)
				{
//					printf("Entity %d is sleeping\n", i);
					body->sleep = true;
					break;
				}
				continue;
			}
			current_time = target_time;
			target_time = delta_time;
		}
		body->net_force = vec3(0.0f, 0.0f, 0.0f);
	}
}


/*
	Handles all collision detection
	return true if simulated too far.
	return false if collision handled.
*/
bool Engine::collision_detect(RigidBody &body)
{
	plane_t floor = { {0.0f, 1.0f, 0.0f}, 3.4f };
	plane_t roof = { {0.0f, -1.0f, 0.0f}, 10.0f };
	plane_t left = { {1.0f, 0.0f, 0.0f}, 10.0f };
	plane_t right = { {-1.0f, 0.0f, 0.0f}, 10.0f };
	plane_t front = { {0.0f, 0.0f, 1.0f}, 25.0f };
	plane_t back = { {0.0f, 0.0f, -1.0f}, 0.0f };

	if (body_collision(body))
		return true;
	else if (body.collision_detect(floor))
		return true;
	else if (body.collision_detect(roof))
		return true;
	else if (body.collision_detect(left))
		return true;
	else if (body.collision_detect(right))
		return true;
	else if (body.collision_detect(front))
		return true;
	else if (body.collision_detect(back))
		return true;
	else
		return false;
}

//O(N^2)
bool Engine::body_collision(RigidBody &body)
{
	for(unsigned int i = 0; i < entity_list.size(); i++)
	{
		if (entity_list[i] == &body)
			continue;

		if ( body.collision_detect(*entity_list[i]) )
			return true;
	}
	return false;
}

bool Engine::mousepos(int x, int y, int deltax, int deltay)
{

	if (keyboard.leftbutton)
	{
		if (keyboard.control)
			light_frame.update(vec2((float)deltax, (float)deltay));
		else
			camera_frame.update(vec2((float)deltax, (float)deltay));
		return true;
	}
	return false;
}

void Engine::keypress(char *key, bool pressed)
{
	char k = 0;

	if (strcmp("enter", key) == 0)
	{
		keyboard.enter = pressed;
	}
	else if (strcmp("leftbutton", key) == 0)
	{
		keyboard.leftbutton = pressed;
		k = 14;
	}
	else if (strcmp("middlebutton", key) == 0)
	{
		keyboard.middlebutton = pressed;
		k = 15;
	}
	else if (strcmp("rightbutton", key) == 0)
	{
		keyboard.rightbutton = pressed;
		k = 16;
	}
	else if (strcmp("shift", key) == 0)
	{
		keyboard.shift = pressed;
	}
	else if (strcmp("control", key) == 0)
	{
		keyboard.control = pressed;
	}
	else if (strcmp("escape", key) == 0)
	{
		keyboard.escape = pressed;
	}
	else if (strcmp("up", key) == 0)
	{
		keyboard.up = pressed;
		k = 3;
	}
	else if (strcmp("left", key) == 0)
	{
		keyboard.left = pressed;
		k = 4;
	}
	else if (strcmp("down", key) == 0)
	{
		keyboard.down = pressed;
		k = 5;
	}
	else if (strcmp("right", key) == 0)
	{
		keyboard.right = pressed;
		k = 6;
	}
}

void Engine::step()
{
	static int time = 0;
	
	if (time == 120)
	{
		RigidBody	*box = new RigidBody;
		box->position = vec3(0.0f, 6.0f, -10.0f);
		box->aabb[0] = vec3(-0.25f, -0.25f, -0.25f);
		box->aabb[1] = vec3(-0.25f, -0.25f, 0.25f);
		box->aabb[2] = vec3(-0.25f, 0.25f, -0.25f);
		box->aabb[3] = vec3(-0.25f, 0.25f, 0.25f);
		box->aabb[4] = vec3(0.25f, -0.25f, -0.25f);
		box->aabb[5] = vec3(0.25f, -0.25f, 0.25f);
		box->aabb[6] = vec3(0.25f, 0.25f, -0.25f);
		box->aabb[7] = vec3(0.25f, 0.25f, 0.25f);
//		box->angular_velocity = vec3(4.0f, 8.0f, 3.0f);
		box->velocity = vec3(-3.0f, 0.0f, -3.0f);
		entity_list.push_back(box);
	}
	time++;


	plane_t plane;

	static int t = 0;

	plane.normal = vec3(sinf(t * 0.1f), cosf(t * 0.1f), sinf(t * 0.1f) + cosf(t * 0.01f));
	plane.normal.normalize();
	plane.d = 0.3f * cosf(t * 0.2f);

	model.load_clipped(gfx, "bunny3", plane);
	t++;

	dynamics();
}


void Engine::resize(int width, int height)
{
	Engine::width = width;
	Engine::height = height;

	//	gfx.resize(width, height);
	glViewport(0, 0, width, height);
	projection.perspective( 45.0, (float) width / height, zNear, zFar, false );
}

void Engine::destroy()
{
	printf("Shutting down.\n");
	gfx.destroy();
//	audio.destroy();
}

void Engine::quit()
{
#ifdef _WINDOWS_
	HWND hwnd = *((HWND *)param1);
	PostMessage(hwnd, WM_CLOSE, 0, 0);
#endif
}
